(use-modules ((srfi srfi-1)  :select (list-index)))
(use-modules ((srfi srfi-13) :select (string-join)))

(define process-arguments
  (let ()
    (define (comma? s) (string=? s ","))
    (define (optional? x) (eq? x #:optional))
    (define (usage program-name arg-handlers)
      (with-output-to-port (current-error-port)
        (lambda ()
          (display "usage: ")
          (display program-name)
          (for-each
           (lambda (handler-entry)
             (display " [")
             (display (car handler-entry))
             (let* ((params (cddr handler-entry))
                    (optional (list-index optional? params))
                    (needed-params (if optional
                                     (list-head optional params)
                                     params))
                    (optional-params (and optional
                                          (cdr (list-tail optional params)))))
               (display (string-join needed-params " " 'prefix))
               (if (and optional-params (not (null? optional-params)))
                 (begin
                   (display " [")
                   (display (string-join optional-params))
                   (display "]")))
               (display " ,]")))
           arg-handlers)
          (newline)))
      (exit 100))
    (define (process-arguments program-name args arg-handlers)
      "Process argument list ARGS according to specification ARG-HANDLERS."
      (let loop ((args args)
                 (apps '()))
        (if (null? args)
          (for-each (lambda (app) (apply (car app) (cdr app))) apps)
          (let* ((handler-entry (assoc (car args) arg-handlers))
                 (dummy (if (not handler-entry)
                          (usage program-name arg-handlers)))
                 (handler (cadr handler-entry))
                 (handler-params (cddr handler-entry))
                 (handler-params-length (length handler-params))
                 (handler-optional (list-index optional? handler-params))
                 (handler-min-arity (or handler-optional
                                        handler-params-length))
                 (handler-max-arity (if handler-optional
                                      (1- handler-params-length)
                                      handler-params-length))
                 (args-rest (cdr args))
                 (args-length (list-index comma? args-rest))
                 (dummy (if (or (not args-length)
                                (< args-length handler-min-arity)
                                (> args-length handler-max-arity))
                          (usage program-name arg-handlers)))
                 (handler-args (list-head args-rest args-length))
                 (args-tail (cdr (list-tail args-rest args-length))))
            (loop args-tail
                  (cons (cons handler handler-args) apps))))))
    process-arguments))
